# Copyright (c) 2021 ING Wholesale Banking Advanced Analytics
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""Monitor thread."""
import http
import threading
import time
import urllib
from typing import Any
from typing import Callable
from typing import List
from typing import Optional
from typing import Union

import requests
import urllib3
from pyspark.sql import SparkSession

import sparkmon


class SparkMon(threading.Thread):
    """Class to monitor a sparkmon application or a Spark Session, running in the background.

    There are multiple design patterns possible for this class.
    One design possibility is to use 2 threads:
    - a daemon thread for the regular update (non-blocking the exit)
    - a non-daemon thread to run the callbacks (blocking the exit)
    Indeed, in this design, it is important to not run the callbacks in the daemon,
    so that a file export callbacks wouldn't be interrupted in the middle of saving.
    Here are the advantage and disadvantage:
    + you can run the callbacks at a slower pace
    - it complexifies at lot and it creates a lot problem like race conditions and dead lock.
    (we had this design at version 0.0.4)

    This is why we took the decision to use only 1 non-daemon thread and check the MainThread status,
    to smoothly stop the monitoring at exit.

    Remark: The same 'application' should not be updated by something else, like other SparkMon instances,
    because it could create race conditions.
    """

    def __init__(
        self,
        application_or_spark: Union[sparkmon.Application, SparkSession],
        period: int = 20,
        callbacks: Optional[List[Callable[..., Any]]] = None,
    ) -> None:
        """Constructor, initializes base class Thread."""
        threading.Thread.__init__(self)
        self.stop_event = threading.Event()
        self.update_cnt = 0
        if isinstance(application_or_spark, sparkmon.Application):
            self.application = application_or_spark
        elif isinstance(application_or_spark, SparkSession):
            self.application = sparkmon.application.create_application_from_spark(application_or_spark)
        else:
            raise TypeError(f"First argument `application_or_spark` is unsupported type {type(application_or_spark)}")
        self.application_lock = threading.Lock()
        self.period = period
        if callbacks is None:
            callbacks = []
        self.callbacks = callbacks
        self.updateEvent = threading.Event()
        self.timeout_sec = 20

    def stop(self) -> None:
        """Don't continue to run the loop, and exit safely the thread."""
        self.stop_event.set()

    def stopped(self) -> bool:
        """Check if we need to stop."""
        return self.stop_event.isSet()

    def is_main_thread_alive(self) -> bool:
        """Check if the main thread is alive."""
        for t in threading.enumerate():
            if t.name == "MainThread":
                return t.is_alive()

    def run(self) -> None:
        """Overrides Thread method."""
        self.start_time = time.time()

        while True:
            if self.stopped():
                return

            # This is a Thread class (non daemon) meaning it can run for ever and block the exit of Python at the end.
            # This is why we check if the main thread is finished to stop SparkMon in a smooth manner at exit:
            if not self.is_main_thread_alive():
                self.stop()

            # We wait at the beginning of the loop, and not at the end,
            # to let the Spark session start and MLflow run at the beginning
            time.sleep(self.period)

            # Updating the application DB
            try:
                # Callbacks are reading application, so let's make thread safe with a lock:
                with self.application_lock:
                    self.application.log_all()
                    self.update_cnt += 1
            except (
                urllib.error.URLError,
                requests.exceptions.ConnectionError,
                http.client.RemoteDisconnected,
                urllib3.exceptions.ProtocolError,
            ) as ex:
                # If we get a connection exception, it either means:
                # - The Spark application didn't fully started yet, and we should wait.
                # - The Spark application is closed, and we should stop monitoring.
                # So we get this exception after the timeout time, let's stop monitoring,
                # so that the main Python process can exit smoothly
                elapsed_sec = time.time() - self.start_time
                if elapsed_sec > self.timeout_sec:
                    # Not need to print if we exited or stopped
                    if not self.stopped():
                        print(
                            f"sparkmon: Info, Spark application not available anymore, stopping monitoring. (Exception: {ex})"
                        )
                    self.stop()
                    return

            # Run the callback
            self.callbacks_run()

            if self.stopped():
                return

            self.updateEvent.set()

    def callbacks_run(self):
        """Running the callbacks."""
        for callback in self.callbacks:
            with self.application_lock:
                callback(self.application)

    def live_plot_notebook(self, n_iter=None) -> None:
        """Useful in the remote case only.

        This might not be compatible with callbacks that are using matplotlib, because matplotlib is not thread safe,
        and you can get the following errors:
        ```
        python(81469,0x1106c5e00) malloc: Incorrect checksum for freed object 0x7fe18da140a8: probably modified after being freed.
        Corrupt value: 0x230017000b00f005
        python(81469,0x1106c5e00) malloc: *** set a breakpoint in malloc_error_break to debug
        ```
        """
        cnt = 0
        while True:
            cnt += 1
            if n_iter is not None and cnt > n_iter:
                return
            if self.stopped():
                return

            with self.application_lock:
                sparkmon.plot_notebook(self.application)
            self.updateEvent.clear()
            self.updateEvent.wait()

    def __enter__(self):
        """Start thread in contextmanager."""
        self.start()
        return self

    def __exit__(self, *args, **kwargs):
        """Stop thread in contextmanager."""
        self.stop()



# Copyright (c) 2021 ING Wholesale Banking Advanced Analytics
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""Utilities functions."""
import collections
import getpass
import math
import os
from typing import Any

import psutil


def convert_size(size_bytes) -> str:
    """Convert bytes in human readble string.

    Credit: https://stackoverflow.com/a/14822210/1876485
    """
    if size_bytes == 0:
        return "0"
    elif size_bytes < 0:
        sign = -1
        size_bytes = size_bytes * -1
    else:
        sign = 1
    size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
    i = int(math.floor(math.log(size_bytes, 1024)))
    p = math.pow(1024, i)
    s = round(size_bytes / p, 2) * sign
    return "%s %s" % (s, size_name[i])


def get_memory_process() -> Any:
    """Get memory usage in bytes of this process running Python."""
    process = psutil.Process(os.getpid())
    return process.memory_info().rss  # in bytes


def get_memory_user() -> Any:
    """Get memory usage in bytes of all process of the current user."""
    user = getpass.getuser()
    # print(user)

    total = 0
    for p in psutil.process_iter():
        try:
            if p.username() == user:
                try:
                    total += p.memory_info().rss
                except psutil.ZombieProcess:
                    # For some process we might get this issue:
                    # psutil.ZombieProcess: psutil.ZombieProcess process still exists but it's a zombie
                    pass
        except psutil.AccessDenied:
            # For some process we might permission issue, with the following error
            # psutil AccessDenied: psutil.AccessDenied (pid=599, name='PanGPS')
            pass

    return total


def flatten_dict(d, parent_key="", sep="_"):
    """Flatten dict recursively.

    Credit: https://stackoverflow.com/a/6027615/1876485
    """
    items = []
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, collections.MutableMapping):
            items.extend(flatten_dict(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)



# Copyright (c) 2021 ING Wholesale Banking Advanced Analytics
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""mlflow missing utilities."""
import tempfile
from contextlib import contextmanager
from pathlib import Path


@contextmanager
def log_file(artifact_full_path: str):
    """Yield a file-object that is going to be saved as an artifact in mlflow.

    mlflow API is really missing this functionality, so let's implement it via a temporary directory.
    """
    import mlflow

    artifact_full_path = Path(artifact_full_path)

    # The artifact_path argument of log_artifact() is actually the directory path
    artifact_path = artifact_full_path.parent
    if len(artifact_path.parts) == 0:
        artifact_path = None
    else:
        artifact_path = str(artifact_path)

    # Let's create a temporary directory and save the file with the same file name
    tmpdir = tempfile.TemporaryDirectory()
    local_path_tmp = Path(tmpdir.name) / artifact_full_path
    local_path_tmp.parent.mkdir(parents=True, exist_ok=True)
    fp = open(local_path_tmp, "w")

    try:
        yield fp
    finally:
        fp.close()
        mlflow.log_artifact(local_path=str(local_path_tmp), artifact_path=artifact_path)

    tmpdir.cleanup()


def active_run():
    """Get the active run with all logs updated."""
    import mlflow

    active_run = mlflow.active_run()
    # active_run.data.params # This is not updated
    return mlflow.get_run(active_run.info.run_id)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['sparkmon']

package_data = \
{'': ['*']}

install_requires = \
['click',
 'data-science-types',
 'ipython',
 'matplotlib',
 'pandas',
 'pandas-stubs',
 'psutil',
 'pyspark',
 'urlpath']

extras_require = \
{'mlflow': ['mlflow']}

entry_points = \
{'console_scripts': ['sparkmon = sparkmon.__main__:main']}

setup_kwargs = {
    'name': 'sparkmon',
    'version': '0.1.5',
    'description': 'sparkmon',
    'long_description': "sparkmon\n========\n\n|PyPI| |Python Version| |License|\n\n|Read the Docs| |Tests| |Codecov|\n\n|pre-commit| |Black|\n\n.. |PyPI| image:: https://img.shields.io/pypi/v/sparkmon.svg\n   :target: https://pypi.org/project/sparkmon/\n   :alt: PyPI\n.. |Python Version| image:: https://img.shields.io/pypi/pyversions/sparkmon\n   :target: https://pypi.org/project/sparkmon\n   :alt: Python Version\n.. |License| image:: https://img.shields.io/pypi/l/sparkmon\n   :target: https://opensource.org/licenses/MIT\n   :alt: License\n.. |Read the Docs| image:: https://img.shields.io/readthedocs/sparkmon/latest.svg?label=Read%20the%20Docs\n   :target: https://sparkmon.readthedocs.io/\n   :alt: Read the documentation at https://sparkmon.readthedocs.io/\n.. |Tests| image:: https://github.com/stephanecollot/sparkmon/workflows/Tests/badge.svg\n   :target: https://github.com/stephanecollot/sparkmon/actions?workflow=Tests\n   :alt: Tests\n.. |Codecov| image:: https://codecov.io/gh/stephanecollot/sparkmon/branch/master/graph/badge.svg\n   :target: https://codecov.io/gh/stephanecollot/sparkmon\n   :alt: Codecov\n.. |pre-commit| image:: https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white\n   :target: https://github.com/pre-commit/pre-commit\n   :alt: pre-commit\n.. |Black| image:: https://img.shields.io/badge/code%20style-black-000000.svg\n   :target: https://github.com/psf/black\n   :alt: Black\n\nDescription\n-----------\n\n``sparkmon`` is a Python package to monitor Spark applications. You can see it as an advanced Spark UI, that keeps track all of `Spark REST API <SparkREST_>`_ metrics **over time**, which makes it quite unique compare to other solutions (see comparison_ below). It is specifically useful to do memory profiling, including Python UDF memory.\n\n\nFeatures\n--------\n\nMonitoring plot example:\n\n.. image:: docs/_static/monitoring-plot-example.png\n\nDisclaimer: Be aware that if you run Spark in local mode some of the subplots will be empty, sparkmon is designed to analyse Spark applications running in a cluster.\n\n* Log the executors metrics\n* Plot monitoring, display in a notebook, or export to a file\n* Can monitor remote Spark application\n* Can run directly in your PySpark application, or run in a notebook, or via the command-line interface\n* Log to mlflow\n\n\nComparison with other solutions\n-------------------------------\n\nThis package brings much more information than Spark UI or other packages. Here is a quick comparison:\n\n- sparkmonitor_:\n\n  - Nice integration in notebook\n  - Doesn't bring more information that Spark UI, specially not memory usage over time.\n\n- sparklint_:\n\n  - Need to launch a server locally, might be difficult on-premise. sparkmon doesn't need to have a port accessible.\n  - Monitors only CPU over time, sparkmon monitors everything including Java and Python memory overtime.\n  - No update since 2018\n\n- `Data Mechanics Delight`_:\n\n  - Really nice and complete\n  - But cannot work fully on-premise\n  - Is not fully open-source\n\n- Sparklens_:\n\n  - But cannot work fully on-premise\n  - Is not fully open-source\n\n\n\nRequirements\n------------\n\n* Python\n* Spark\n* mlflow (optional)\n\n\nInstallation\n------------\n\nYou can install *sparkmon* via pip_ from PyPI_:\n\n.. code:: console\n\n   $ pip install sparkmon\n   $ pip install sparkmon[mlflow]\n\n\nUsage\n-----\n\nSimple use-case:\n\n.. code-block:: python\n\n   import sparkmon\n\n   # Create and start the monitoring process via a Spark session\n   mon = sparkmon.SparkMon(spark, period=5, callbacks=[\n       sparkmon.callbacks.plot_to_image,\n       sparkmon.callbacks.log_to_mlflow,\n   ])\n   mon.start()\n\n   # Stop monitoring\n   mon.stop()\n\nMore advanced use-case:\n\n.. code-block:: python\n\n   import sparkmon\n\n   # Create an app connection\n   # via a Spark session\n   application = sparkmon.create_application_from_spark(spark)\n   # or via a remote Spark web UI link\n   application = sparkmon.create_application_from_link(index=0, web_url='http://localhost:4040')\n\n   # Create and start the monitoring process\n   mon = sparkmon.SparkMon(application, period=5, callbacks=[\n       sparkmon.callbacks.plot_to_image,\n       sparkmon.callbacks.log_to_mlflow,\n   ])\n   mon.start()\n\n   # Stop monitoring\n   mon.stop()\n\nYou can also use it from a notebook: `Notebook Example <Example_>`_\n\nThere is also a command-line interface, see  `Command-line Reference <Usage_>`_ for details.\n\n\nHow does it work?\n-----------------\n\n``SparkMon`` is running in the background a Python thread that is querying Spark web UI API and logging all the executors information over time.\n\nThe ``callbacks`` list parameters allows you to define what do after each update, like exporting executors historical info to a csv, or plotting to a file, or to your notebook.\n\n\nContributing\n------------\n\nContributions are very welcome.\nTo learn more, see the `Contributor Guide`_.\n\n\nLicense\n-------\n\nDistributed under the terms of the `MIT license`_,\n*sparkmon* is free and open source software.\n\n\nIssues\n------\n\nIf you encounter any problems,\nplease `file an issue`_ along with a detailed description.\n\n\nCredits\n-------\n\nThis project was generated from `@cjolowicz`_'s `Hypermodern Python Cookiecutter`_ template.\n\n.. _@cjolowicz: https://github.com/cjolowicz\n.. _Cookiecutter: https://github.com/audreyr/cookiecutter\n.. _MIT license: https://opensource.org/licenses/MIT\n.. _PyPI: https://pypi.org/\n.. _Hypermodern Python Cookiecutter: https://github.com/cjolowicz/cookiecutter-hypermodern-python\n.. _file an issue: https://github.com/stephanecollot/sparkmon/issues\n.. _pip: https://pip.pypa.io/\n.. github-only\n.. _Contributor Guide: CONTRIBUTING.rst\n.. _Usage: https://sparkmon.readthedocs.io/en/latest/usage.html\n.. _Example: https://sparkmon.readthedocs.io/en/latest/example.html\n.. _SparkREST: https://spark.apache.org/docs/latest/monitoring.html#rest-api\n.. _sparkmonitor: https://krishnan-r.github.io/sparkmonitor/\n.. _sparklint: https://github.com/groupon/sparklint\n.. _comparison: #comparison-with-other-solutions\n.. _Data Mechanics Delight: https://www.datamechanics.co/delight\n.. _Sparklens: http://sparklens.qubole.com/\n",
    'author': 'Stephane Collot',
    'author_email': 'stephane.collot@ing.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/stephanecollot/sparkmon',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)
